"""This module contains a representation of the EVM's machine state and its
stack."""
from copy import copy
from typing import cast, Sized, Union, Any, List, Dict, Optional

from mythril.laser.smt import BitVec, Bool, If, Expression, symbol_factory

from ethereum import opcodes, utils
from mythril.laser.ethereum.evm_exceptions import (
    StackOverflowException,
    StackUnderflowException,
    OutOfGasException,
)
from mythril.laser.ethereum.state.memory import Memory


class MachineStack(list):
    """Defines EVM stack, overrides the default list to handle overflows."""

    STACK_LIMIT = 1024

    def __init__(self, default_list=None) -> None:
        """

        :param default_list:
        """
        super(MachineStack, self).__init__(default_list or [])

    def append(self, element: Union[int, Expression]) -> None:
        """
        This function ensures the following properties when appending to a list:
            - Element appended to this list should be a BitVec
            - Ensures stack overflow bound
        :param element: element to be appended to the list
        :function: appends the element to list if the size is less than STACK_LIMIT, else throws an error
        """
        if isinstance(element, int):
            element = symbol_factory.BitVecVal(element, 256)
        if isinstance(element, Bool):
            element = If(
                element,
                symbol_factory.BitVecVal(1, 256),
                symbol_factory.BitVecVal(0, 256),
            )
        if super(MachineStack, self).__len__() >= self.STACK_LIMIT:
            raise StackOverflowException(
                "Reached the EVM stack limit of {}, you can't append more "
                "elements".format(self.STACK_LIMIT)
            )
        super(MachineStack, self).append(element)

    def pop(self, index=-1) -> Union[int, Expression]:
        """
        This function ensures stack underflow bound
        :param index:index to be popped, same as the list() class.
        :returns popped value
        :function: same as list() class but throws StackUnderflowException for popping from an empty list
        """

        try:
            return super(MachineStack, self).pop(index)
        except IndexError:
            raise StackUnderflowException("Trying to pop from an empty stack")

    def __getitem__(self, item: Union[int, slice]) -> Any:
        """

        :param item:
        :return:
        """
        try:
            return super(MachineStack, self).__getitem__(item)
        except IndexError:
            raise StackUnderflowException(
                "Trying to access a stack element which doesn't exist"
            )

    def __add__(self, other):
        """Implement list concatenation if needed.

        :param other:
        """
        raise NotImplementedError("Implement this if needed")

    def __iadd__(self, other):
        """Implement list concatenation if needed.

        :param other:
        """
        raise NotImplementedError("Implement this if needed")


class MachineState:
    """
        MachineState represents current machine state also referenced to as \mu.
    """

    def __init__(
        self,
        gas_limit: int,
        pc=0,
        stack=None,
        subroutine_stack=None,
        memory: Optional[Memory] = None,
        constraints=None,
        depth=0,
        max_gas_used=0,
        min_gas_used=0,
        prev_pc=-1,
    ) -> None:
        """Constructor for machineState.

        :param gas_limit:
        :param pc:
        :param stack:
        :param memory:
        :param constraints:
        :param depth:
        :param max_gas_used:
        :param min_gas_used:
        :param prev_pc:
        """
        self._pc = pc
        self.stack = MachineStack(stack)
        self.subroutine_stack = MachineStack(subroutine_stack)
        self.memory = memory or Memory()
        self.gas_limit = gas_limit
        self.min_gas_used = min_gas_used  # lower gas usage bound
        self.max_gas_used = max_gas_used  # upper gas usage bound
        self.depth = depth
        self.prev_pc = prev_pc  # holds context of current pc

    def calculate_extension_size(self, start: int, size: int) -> int:
        """

        :param start:
        :param size:
        :return:
        """
        if self.memory_size > start + size:
            return 0

        # The extension size is calculated based on the word length
        new_size = utils.ceil32(start + size) // 32
        old_size = self.memory_size // 32

        return (new_size - old_size) * 32

    def calculate_memory_gas(self, start: int, size: int):
        """

        :param start:
        :param size:
        :return:
        """
        # https://github.com/ethereum/pyethereum/blob/develop/ethereum/vm.py#L148
        oldsize = self.memory_size // 32
        old_totalfee = (
            oldsize * opcodes.GMEMORY + oldsize ** 2 // opcodes.GQUADRATICMEMDENOM
        )
        newsize = utils.ceil32(start + size) // 32
        new_totalfee = (
            newsize * opcodes.GMEMORY + newsize ** 2 // opcodes.GQUADRATICMEMDENOM
        )
        return new_totalfee - old_totalfee

    def check_gas(self):
        """Check whether the machine is out of gas."""
        if self.min_gas_used > self.gas_limit:
            raise OutOfGasException()

    def mem_extend(self, start: Union[int, BitVec], size: Union[int, BitVec]) -> None:
        """Extends the memory of this machine state.

        :param start: Start of memory extension
        :param size: Size of memory extension
        """
        if (isinstance(start, BitVec) and start.symbolic) or (
            isinstance(size, BitVec) and size.symbolic
        ):
            return
        if isinstance(start, BitVec):
            start = start.value
        if isinstance(size, BitVec):
            size = size.value
        m_extend = self.calculate_extension_size(start, size)
        if m_extend:
            extend_gas = self.calculate_memory_gas(start, size)
            self.min_gas_used += extend_gas
            self.max_gas_used += extend_gas
            self.check_gas()
            self.memory.extend(m_extend)

    def memory_write(self, offset: int, data: List[Union[int, BitVec]]) -> None:
        """Writes data to memory starting at offset.

        :param offset:
        :param data:
        """
        self.mem_extend(offset, len(data))
        self.memory[offset : offset + len(data)] = data

    def pop(self, amount=1) -> Union[BitVec, List[BitVec]]:
        """Pops amount elements from the stack.

        :param amount:
        :return:
        """
        if amount > len(self.stack):
            raise StackUnderflowException
        values = self.stack[-amount:][::-1]
        del self.stack[-amount:]

        return values[0] if amount == 1 else values

    def __deepcopy__(self, memodict=None):
        """

        :param memodict:
        :return:
        """
        memodict = {} if memodict is None else memodict
        return MachineState(
            gas_limit=self.gas_limit,
            max_gas_used=self.max_gas_used,
            min_gas_used=self.min_gas_used,
            pc=self._pc,
            stack=copy(self.stack),
            memory=copy(self.memory),
            depth=self.depth,
            prev_pc=self.prev_pc,
            subroutine_stack=copy(self.subroutine_stack),
        )

    def __str__(self):
        """

        :return:
        """
        return str(self.as_dict)

    @property
    def pc(self) -> int:
        """

        :return:
        """
        return self._pc

    @pc.setter
    def pc(self, value):
        self.prev_pc = self._pc
        self._pc = value

    @property
    def memory_size(self) -> int:
        """

        :return:
        """
        return len(cast(Sized, self.memory))

    @property
    def as_dict(self) -> Dict:
        """

        :return:
        """
        return dict(
            pc=self._pc,
            stack=self.stack,
            subroutine_stack=self.subroutine_stack,
            memory=self.memory,
            memsize=self.memory_size,
            gas=self.gas_limit,
            max_gas_used=self.max_gas_used,
            min_gas_used=self.min_gas_used,
            prev_pc=self.prev_pc,
        )



from abc import ABC, abstractmethod
from mythril.laser.plugin.builder import PluginBuilder as LaserPluginBuilder


class MythrilPlugin:
    """MythrilPlugin interface

    Mythril Plugins can be used to extend Mythril in different ways:
    1. Extend Laser, in which case the LaserPlugin interface must also be extended
    2. Extend Laser with a new search strategy in which case the SearchStrategy needs to be implemented
    3. Add an analysis module, in this case the AnalysisModule interface needs to be implemented
    4. Add new commands to the Mythril cli, using the MythrilCLIPlugin Interface
    """

    author = "Default Author"
    name = "Plugin Name"
    plugin_license = "All rights reserved."
    plugin_type = "Mythril Plugin"
    plugin_version = "0.0.1 "
    plugin_description = "This is an example plugin description"

    def __init__(self, **kwargs):
        pass

    def __repr__(self):
        plugin_name = type(self).__name__
        return f"{plugin_name} - {self.plugin_version} - {self.author}"


class MythrilCLIPlugin(MythrilPlugin):
    """MythrilCLIPlugin interface

    This interface should be implemented by mythril plugins that aim to add commands to the mythril cli
    """

    pass


class MythrilLaserPlugin(MythrilPlugin, LaserPluginBuilder, ABC):
    """ Mythril Laser Plugin interface

    Plugins of this type are used to instrument the laser EVM
    """

    pass



""" Mythril Detection Modules

This module includes an definition of the DetectionModule interface.
DetectionModules implement different analysis rules to find weaknesses and vulnerabilities.
"""
import logging
from typing import List, Set, Optional, Tuple, Union

from mythril.analysis.report import Issue
from mythril.laser.ethereum.state.global_state import GlobalState

from abc import ABC, abstractmethod
from enum import Enum

# Get logger instance
log = logging.getLogger(__name__)


class EntryPoint(Enum):
    """ EntryPoint Enum

    This enum is used to signify the entry_point of detection modules.
    See also the class documentation of DetectionModule
    """

    POST = 1
    CALLBACK = 2


class DetectionModule(ABC):
    """The base detection module.

    All custom-built detection modules must inherit from this class.

    There are several class properties that expose information about the detection modules
    - name: The name of the detection module
    - swc_id: The SWC ID associated with the weakness that the module detects
    - description: A description of the detection module, and what it detects
    - entry_point: Mythril can run callback style detection modules, or modules that search the statespace.
                [IMPORTANT] POST entry points severely slow down the analysis, try to always use callback style modules
    - pre_hooks: A list of instructions to hook the laser vm for (pre execution of the instruction)
    - post_hooks: A list of instructions to hook the laser vm for (post execution of the instruction)
    """

    name = "Detection Module Name / Title"
    swc_id = "SWC-000"
    description = "Detection module description"
    entry_point = EntryPoint.CALLBACK  # type: EntryPoint
    pre_hooks = []  # type: List[str]
    post_hooks = []  # type: List[str]

    def __init__(self) -> None:
        self.issues = []  # type: List[Issue]
        self.cache = set()  # type: Set[Optional[Union[int, Tuple[int, str]]]]

    def reset_module(self):
        """ Resets the storage of this module """
        self.issues = []

    def execute(self, target: GlobalState) -> Optional[List[Issue]]:
        """The entry point for execution, which is being called by Mythril.

        :param target: The target of the analysis, either a global state (callback) or the entire statespace (post)
        :return: List of encountered issues
        """

        log.debug("Entering analysis module: {}".format(self.__class__.__name__))

        result = self._execute(target)

        log.debug("Exiting analysis module: {}".format(self.__class__.__name__))

        return result

    @abstractmethod
    def _execute(self, target) -> Optional[List[Issue]]:
        """Module main method (override this)

        :param target: The target of the analysis, either a global state (callback) or the entire statespace (post)
        :return: List of encountered issues
        """
        pass

    def __repr__(self) -> str:
        return (
            "<"
            "DetectionModule "
            "name={0.name} "
            "swc_id={0.swc_id} "
            "pre_hooks={0.pre_hooks} "
            "post_hooks={0.post_hooks} "
            "description={0.description}"
            ">"
        ).format(self)



"""This module contains EVM exception types used by LASER."""


class VmException(Exception):
    """The base VM exception type."""

    pass


class StackUnderflowException(IndexError, VmException):
    """A VM exception regarding stack underflows."""

    pass


class StackOverflowException(VmException):
    """A VM exception regarding stack overflows."""

    pass


class InvalidJumpDestination(VmException):
    """A VM exception regarding JUMPs to invalid destinations."""

    pass


class InvalidInstruction(VmException):
    """A VM exception denoting an invalid op code has been encountered."""

    pass


class OutOfGasException(VmException):
    """A VM exception denoting the current execution has run out of gas."""

    pass


class WriteProtection(VmException):
    """A VM exception denoting that a write operation is executed on a write protected environment"""

    pass



"""This module contains the detection code for reachable exceptions."""
import logging

from typing import List, cast
from mythril.analysis import solver
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.analysis.report import Issue
from mythril.analysis.swc_data import ASSERT_VIOLATION
from mythril.exceptions import UnsatError
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.laser.ethereum import util

log = logging.getLogger(__name__)

# The function signature of Panic(uint256)
PANIC_SIGNATURE = [78, 72, 123, 113]


class LastJumpAnnotation(StateAnnotation):
    """ State Annotation used if an overflow is both possible and used in the annotated path"""

    def __init__(self, last_jump: int = None) -> None:
        self.last_jump: int = last_jump

    def __copy__(self):
        new_annotation = LastJumpAnnotation(self.last_jump)
        return new_annotation


class Exceptions(DetectionModule):
    """"""

    name = "Assertion violation"
    swc_id = ASSERT_VIOLATION
    description = "Checks whether any exception states are reachable."
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["INVALID", "JUMP", "REVERT"]

    def _execute(self, state: GlobalState) -> None:
        """

        :param state:
        :return:
        """
        if (
            state.get_current_instruction()["address"],
            state.environment.active_function_name,
        ) in self.cache:
            return
        issues = self._analyze_state(state)
        for issue in issues:
            self.cache.add((issue.address, issue.function))
        self.issues.extend(issues)

    @staticmethod
    def _analyze_state(state) -> List[Issue]:
        """

        :param state:
        :return:
        """
        opcode = state.get_current_instruction()["opcode"]
        address = state.get_current_instruction()["address"]

        annotations = cast(
            List[LastJumpAnnotation],
            [a for a in state.get_annotations(LastJumpAnnotation)],
        )
        if len(annotations) == 0:
            state.annotate(LastJumpAnnotation())
            annotations = cast(
                List[LastJumpAnnotation],
                [a for a in state.get_annotations(LastJumpAnnotation)],
            )

        if opcode == "JUMP":
            annotations[0].last_jump = address
            return []

        if opcode == "REVERT" and not is_assertion_failure(state):
            return []

        log.debug(
            "ASSERT_FAIL/REVERT in function " + state.environment.active_function_name
        )

        try:
            description_tail = (
                "It is possible to trigger an assertion violation. Note that Solidity assert() statements should "
                "only be used to check invariants. Review the transaction trace generated for this issue and "
                "either make sure your program logic is correct, or use require() instead of assert() if your goal "
                "is to constrain user inputs or enforce preconditions. Remember to validate inputs from both callers "
                "(for instance, via passed arguments) and callees (for instance, via return values)."
            )
            transaction_sequence = solver.get_transaction_sequence(
                state, state.world_state.constraints
            )

            issue = Issue(
                contract=state.environment.active_account.contract_name,
                function_name=state.environment.active_function_name,
                address=address,
                swc_id=ASSERT_VIOLATION,
                title="Exception State",
                severity="Medium",
                description_head="An assertion violation was triggered.",
                description_tail=description_tail,
                bytecode=state.environment.code.bytecode,
                transaction_sequence=transaction_sequence,
                gas_used=(state.mstate.min_gas_used, state.mstate.max_gas_used),
                source_location=annotations[0].last_jump,
            )
            return [issue]

        except UnsatError:
            log.debug("no model found")

        return []


def is_assertion_failure(global_state):
    state = global_state.mstate
    offset, length = state.stack[-1], state.stack[-2]
    try:
        return_data = state.memory[
            util.get_concrete_int(offset) : util.get_concrete_int(offset + length)
        ]
    except TypeError:
        return False

    return return_data[:4] == PANIC_SIGNATURE and return_data[-1] == 1


detector = Exceptions()



""" Plugin implementations

This module contains the implementation of some features

- benchmarking
- pruning
"""
from mythril.laser.plugin.plugins.benchmark import BenchmarkPluginBuilder
from mythril.laser.plugin.plugins.coverage.coverage_plugin import CoveragePluginBuilder
from mythril.laser.plugin.plugins.dependency_pruner import DependencyPrunerBuilder
from mythril.laser.plugin.plugins.mutation_pruner import MutationPrunerBuilder
from mythril.laser.plugin.plugins.call_depth_limiter import CallDepthLimitBuilder
from mythril.laser.plugin.plugins.instruction_profiler import InstructionProfilerBuilder



"""This module contains various utility functions regarding unit conversion and
solc integration."""
import binascii
import json
import sys
import os
import platform
import logging
import solc

from pathlib import Path
from subprocess import PIPE, Popen
from ethereum.abi import encode_abi, encode_int, method_id
from ethereum.utils import zpad
from json.decoder import JSONDecodeError
from mythril.exceptions import CompilerError
from semantic_version import Version

if sys.version_info[1] >= 6:
    import solcx
    from solcx.exceptions import SolcNotInstalled

log = logging.getLogger(__name__)


def safe_decode(hex_encoded_string):
    """

    :param hex_encoded_string:
    :return:
    """
    if hex_encoded_string.startswith("0x"):
        return bytes.fromhex(hex_encoded_string[2:])
    else:
        return bytes.fromhex(hex_encoded_string)


def get_solc_json(file, solc_binary="solc", solc_settings_json=None):
    """

    :param file:
    :param solc_binary:
    :param solc_settings_json:
    :return:
    """
    cmd = [solc_binary, "--standard-json", "--allow-paths", "."]
    settings = {}
    if solc_settings_json:
        with open(solc_settings_json) as f:
            settings = json.load(f)
    settings.update(
        {
            "optimizer": {"enabled": False},
            "outputSelection": {
                "*": {
                    "": ["ast"],
                    "*": [
                        "metadata",
                        "evm.bytecode",
                        "evm.deployedBytecode",
                        "evm.methodIdentifiers",
                    ],
                }
            },
        }
    )

    input_json = json.dumps(
        {
            "language": "Solidity",
            "sources": {file: {"urls": [file]}},
            "settings": settings,
        }
    )

    try:
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate(bytes(input_json, "utf8"))

    except FileNotFoundError:
        raise CompilerError(
            "Compiler not found. Make sure that solc is installed and in PATH, or set the SOLC environment variable."
        )

    out = stdout.decode("UTF-8")

    try:
        result = json.loads(out)
    except JSONDecodeError as e:
        log.error(f"Encountered a decode error, stdout:{out}, stderr: {stderr}")
        raise e

    for error in result.get("errors", []):
        if error["severity"] == "error":
            raise CompilerError(
                "Solc experienced a fatal error.\n\n%s" % error["formattedMessage"]
            )

    return result


def encode_calldata(func_name, arg_types, args):
    """

    :param func_name:
    :param arg_types:
    :param args:
    :return:
    """
    mid = method_id(func_name, arg_types)
    function_selector = zpad(encode_int(mid), 4)
    args = encode_abi(arg_types, args)
    return "0x" + function_selector.hex() + args.hex()


def get_random_address():
    """

    :return:
    """
    return binascii.b2a_hex(os.urandom(20)).decode("UTF-8")


def get_indexed_address(index):
    """

    :param index:
    :return:
    """
    return "0x" + (hex(index)[2:] * 40)


def solc_exists(version):
    """

    :param version:
    :return:
    """

    default_binary = "/usr/bin/solc"
    if sys.version_info[1] >= 6:
        if platform.system() == "Darwin":
            solcx.import_installed_solc()
        solcx.install_solc("v" + version)
        solcx.set_solc_version("v" + version)
        solc_binary = solcx.install.get_executable()
        return solc_binary
    elif Version("0.4.2") <= Version(version) <= Version("0.4.25"):
        if not solc.main.is_solc_available():
            solc.install_solc("v" + version)
            solc_binary = solc.install.get_executable_path("v" + version)
            return solc_binary
        else:
            solc_binaries = [
                os.path.join(
                    os.environ.get("HOME", str(Path.home())),
                    ".py-solc/solc-v" + version,
                    "bin/solc",
                )  # py-solc setup
            ]
            for solc_path in solc_binaries:
                if os.path.exists(solc_path):
                    return solc_path
    elif os.path.exists(default_binary):
        return default_binary

    else:
        return None



from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.strategy.basic import BasicSearchStrategy
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.laser.ethereum.transaction import ContractCreationTransaction
from typing import Dict, cast, List
from copy import copy
import logging


log = logging.getLogger(__name__)


class JumpdestCountAnnotation(StateAnnotation):
    """State annotation that counts the number of jumps per destination."""

    def __init__(self) -> None:
        self._reached_count = {}  # type: Dict[int, int]
        self.trace = []  # type: List[int]

    def __copy__(self):
        result = JumpdestCountAnnotation()
        result._reached_count = copy(self._reached_count)
        result.trace = copy(self.trace)
        return result


class BoundedLoopsStrategy(BasicSearchStrategy):
    """Adds loop pruning to the search strategy.
    Ignores JUMPI instruction if the destination was targeted >JUMPDEST_LIMIT times.
    """

    def __init__(self, super_strategy: BasicSearchStrategy, *args) -> None:
        """"""

        self.super_strategy = super_strategy
        self.bound = args[0][0]

        log.info(
            "Loaded search strategy extension: Loop bounds (limit = {})".format(
                self.bound
            )
        )

        BasicSearchStrategy.__init__(
            self, super_strategy.work_list, super_strategy.max_depth
        )

    @staticmethod
    def calculate_hash(i: int, j: int, trace: List[int]) -> int:
        """
        calculate hash(trace[i: j])
        :param i:
        :param j:
        :param trace:
        :return: hash(trace[i: j])
        """
        key = 0
        size = 0
        for itr in range(i, j):
            key |= trace[itr] << ((itr - i) * 8)
            size += 1

        return key

    @staticmethod
    def count_key(trace: List[int], key: int, start: int, size: int) -> int:
        """
        Count continuous loops in the trace.
        :param trace:
        :param key:
        :param size:
        :return:
        """
        count = 1
        i = start
        while i >= 0:
            if BoundedLoopsStrategy.calculate_hash(i, i + size, trace) != key:
                break
            count += 1
            i -= size
        return count

    @staticmethod
    def get_loop_count(trace: List[int]) -> int:
        """
        Gets the loop count
        :param trace: annotation trace
        :return:
        """
        found = False
        for i in range(len(trace) - 3, 0, -1):
            if trace[i] == trace[-2] and trace[i + 1] == trace[-1]:
                found = True
                break

        if found:
            key = BoundedLoopsStrategy.calculate_hash(i + 1, len(trace) - 1, trace)
            size = len(trace) - i - 2
            count = BoundedLoopsStrategy.count_key(trace, key, i + 1, size)
        else:
            count = 0
        return count

    def get_strategic_global_state(self) -> GlobalState:
        """ Returns the next state

        :return: Global state
        """

        while True:

            state = self.super_strategy.get_strategic_global_state()

            annotations = cast(
                List[JumpdestCountAnnotation],
                list(state.get_annotations(JumpdestCountAnnotation)),
            )

            if len(annotations) == 0:
                annotation = JumpdestCountAnnotation()
                state.annotate(annotation)
            else:
                annotation = annotations[0]

            cur_instr = state.get_current_instruction()

            annotation.trace.append(cur_instr["address"])

            if cur_instr["opcode"].upper() != "JUMPDEST":
                return state

            # create unique instruction identifier
            count = BoundedLoopsStrategy.get_loop_count(annotation.trace)
            # The creation transaction gets a higher loop bound to give it a better chance of success.
            # TODO: There's probably a nicer way to do this
            if isinstance(
                state.current_transaction, ContractCreationTransaction
            ) and count < max(8, self.bound):
                return state

            elif count > self.bound:
                log.debug("Loop bound reached, skipping state")
                continue

            return state



"""This module contains functionality for hooking in detection modules and
executing them."""

from mythril.analysis.module import ModuleLoader, reset_callback_modules
from mythril.analysis.module.base import EntryPoint
from mythril.analysis.report import Issue

from typing import Optional, List
import logging

log = logging.getLogger(__name__)


def retrieve_callback_issues(white_list: Optional[List[str]] = None) -> List[Issue]:
    """ Get the issues discovered by callback type detection modules"""
    issues = []  # type: List[Issue]
    for module in ModuleLoader().get_detection_modules(
        entry_point=EntryPoint.CALLBACK, white_list=white_list
    ):
        log.debug("Retrieving results for " + module.name)
        issues += module.issues

    reset_callback_modules(module_names=white_list)

    return issues


def fire_lasers(statespace, white_list: Optional[List[str]] = None) -> List[Issue]:
    """ Fire lasers at analysed statespace object

    :param statespace: Symbolic statespace to analyze
    :param white_list: Optionally whitelist modules to use for the analysis
    :return: Discovered issues
    """
    log.info("Starting analysis")

    issues = []  # type: List[Issue]
    for module in ModuleLoader().get_detection_modules(
        entry_point=EntryPoint.POST, white_list=white_list
    ):
        log.info("Executing " + module.name)
        issues += module.execute(statespace)

    issues += retrieve_callback_issues(white_list)
    return issues



"""This module."""
from enum import Enum
from typing import Dict, List, TYPE_CHECKING

from mythril.laser.ethereum.state.constraints import Constraints
from flags import Flags

if TYPE_CHECKING:
    from mythril.laser.ethereum.state.global_state import GlobalState


class JumpType(Enum):
    """An enum to represent the types of possible JUMP scenarios."""

    CONDITIONAL = 1
    UNCONDITIONAL = 2
    CALL = 3
    RETURN = 4
    Transaction = 5


class NodeFlags(Flags):
    """A collection of flags to denote the type a call graph node can have."""

    def __or__(self, other) -> "NodeFlags":
        return super().__or__(other)

    FUNC_ENTRY = 1
    CALL_RETURN = 2


class Node:
    """The representation of a call graph node."""

    def __init__(
        self,
        contract_name: str,
        start_addr=0,
        constraints=None,
        function_name="unknown",
    ) -> None:
        """

        :param contract_name:
        :param start_addr:
        :param constraints:
        """
        constraints = constraints if constraints else Constraints()
        self.contract_name = contract_name
        self.start_addr = start_addr
        self.states = []  # type: List[GlobalState]
        self.constraints = constraints
        self.function_name = function_name
        self.flags = NodeFlags()

        self.uid = hash(self)

    def get_cfg_dict(self) -> Dict:
        """

        :return:
        """
        code = ""
        for state in self.states:
            instruction = state.get_current_instruction()

            code += str(instruction["address"]) + " " + instruction["opcode"]
            if instruction["opcode"].startswith("PUSH"):
                code += " " + instruction["argument"]

            code += "\\n"

        return dict(
            contract_name=self.contract_name,
            start_addr=self.start_addr,
            function_name=self.function_name,
            code=code,
        )


class Edge:
    """The respresentation of a call graph edge."""

    def __init__(
        self,
        node_from: int,
        node_to: int,
        edge_type=JumpType.UNCONDITIONAL,
        condition=None,
    ) -> None:
        """

        :param node_from:
        :param node_to:
        :param edge_type:
        :param condition:
        """
        self.node_from = node_from
        self.node_to = node_to
        self.type = edge_type
        self.condition = condition

    def __str__(self) -> str:
        """

        :return:
        """
        return str(self.as_dict)

    @property
    def as_dict(self) -> Dict[str, int]:
        """

        :return:
        """
        return {"from": self.node_from, "to": self.node_to}



from collections import defaultdict
from typing import List, Optional, Callable, Mapping, Dict
import logging

from mythril.support.opcodes import OPCODES
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.analysis.module.loader import ModuleLoader

log = logging.getLogger(__name__)
OP_CODE_LIST = OPCODES.keys()


def get_detection_module_hooks(
    modules: List[DetectionModule], hook_type="pre"
) -> Dict[str, List[Callable]]:
    """ Gets a dictionary with the hooks for the passed detection modules

    :param modules: The modules for which to retrieve hooks
    :param hook_type: The type  of hooks to retrieve (default: "pre")
    :return: Dictionary with discovered hooks
    """
    hook_dict = defaultdict(list)  # type: Mapping[str, List[Callable]]
    for module in modules:

        hooks = module.pre_hooks if hook_type == "pre" else module.post_hooks

        for op_code in map(lambda x: x.upper(), hooks):
            # A hook can be either OP_CODE or START*
            # When an entry like the second is encountered we hook all opcodes that start with START
            if op_code in OP_CODE_LIST:
                hook_dict[op_code].append(module.execute)
            elif op_code.endswith("*"):
                to_register = filter(lambda x: x.startswith(op_code[:-1]), OP_CODE_LIST)
                for actual_hook in to_register:
                    hook_dict[actual_hook].append(module.execute)
            else:
                log.error(
                    "Encountered invalid hook opcode %s in module %s",
                    op_code,
                    module.name,
                )

    return dict(hook_dict)


def reset_callback_modules(module_names: Optional[List[str]] = None):
    """Clean the issue records of every callback-based module."""
    modules = ModuleLoader().get_detection_modules(EntryPoint.CALLBACK, module_names)
    for module in modules:
        module.reset_module()



from mythril.solidity.soliditycontract import SolidityContract
from mythril.ethereum.evmcontract import EVMContract


class Source:
    """Class to handle to source data"""

    def __init__(self, source_type=None, source_format=None, source_list=None):
        """
        :param source_type: whether it is a solidity-file or evm-bytecode
        :param source_format: whether it is bytecode, ethereum-address or text
        :param source_list: List of files
        :param meta: meta data
        """
        self.source_type = source_type
        self.source_format = source_format
        self.source_list = source_list or []
        self._source_hash = []

    def get_source_from_contracts_list(self, contracts):
        """
        get the source data from the contracts list
        :param contracts: the list of contracts
        :return:
        """
        if contracts is None or len(contracts) == 0:
            return
        if isinstance(contracts[0], SolidityContract):
            self.source_type = "solidity-file"
            self.source_format = "text"
            for contract in contracts:
                self.source_list += [
                    file.filename for file in contract.solc_indices.values()
                ]
                self._source_hash.append(contract.bytecode_hash)
                self._source_hash.append(contract.creation_bytecode_hash)
        elif isinstance(contracts[0], EVMContract):
            self.source_format = "evm-byzantium-bytecode"
            self.source_type = (
                "ethereum-address"
                if len(contracts[0].name) == 42 and contracts[0].name[0:2] == "0x"
                else "raw-bytecode"
            )
            for contract in contracts:
                if contract.creation_code:
                    self.source_list.append(contract.creation_bytecode_hash)
                if contract.code:
                    self.source_list.append(contract.bytecode_hash)
            self._source_hash = self.source_list

        else:
            assert False  # Fail hard

    def get_source_index(self, bytecode_hash: str) -> int:
        """
        Find the contract index in the list
        :param bytecode_hash: The contract hash
        :return: The index of the contract in the _source_hash list
        """
        # TODO: Add this part to exception logs
        try:
            return self._source_hash.index(bytecode_hash)
        except ValueError:
            self._source_hash.append(bytecode_hash)
            return len(self._source_hash) - 1



"""This module contains various utility functions regarding the RPC data format
and validation."""
from .constants import BLOCK_TAGS


def hex_to_dec(x):
    """Convert hex to decimal.

    :param x:
    :return:
    """
    return int(x, 16)


def clean_hex(d):
    """Convert decimal to hex and remove the "L" suffix that is appended to
    large numbers.

    :param d:
    :return:
    """
    return hex(d).rstrip("L")


def validate_block(block):
    """

    :param block:
    :return:
    """
    if isinstance(block, str):
        if block not in BLOCK_TAGS:
            raise ValueError("invalid block tag")
    if isinstance(block, int):
        block = hex(block)
    return block


def wei_to_ether(wei):
    """Convert wei to ether.

    :param wei:
    :return:
    """
    return 1.0 * wei / 10 ** 18


def ether_to_wei(ether):
    """Convert ether to wei.

    :param ether:
    :return:
    """
    return ether * 10 ** 18



"""This module contains exceptions regarding JSON-RPC communication."""


class EthJsonRpcError(Exception):
    """The JSON-RPC base exception type."""

    pass


class ConnectionError(EthJsonRpcError):
    """An RPC exception denoting there was an error in connecting to the RPC
    instance."""

    pass


class BadStatusCodeError(EthJsonRpcError):
    """An RPC exception denoting a bad status code returned by the RPC
    instance."""

    pass


class BadJsonError(EthJsonRpcError):
    """An RPC exception denoting that the RPC instance returned a bad JSON
    object."""

    pass


class BadResponseError(EthJsonRpcError):
    """An RPC exception denoting that the RPC instance returned a bad
    response."""

    pass



import pytest
from mock import patch

from mythril.disassembler.disassembly import Disassembly
from mythril.laser.smt import symbol_factory
from mythril.laser.ethereum.state.environment import Environment
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.state.machine_state import MachineState
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.instructions import Instruction
from mythril.laser.ethereum.transaction.transaction_models import MessageCallTransaction
from mythril.laser.ethereum.call import SymbolicCalldata
from mythril.laser.ethereum.transaction import TransactionStartSignal

from mythril.laser.ethereum.evm_exceptions import WriteProtection


def get_global_state():
    active_account = Account("0x0", code=Disassembly("60606040"))
    environment = Environment(
        active_account, None, SymbolicCalldata("2"), None, None, None, None
    )
    world_state = WorldState()
    world_state.put_account(active_account)
    state = GlobalState(world_state, environment, None, MachineState(gas_limit=8000000))
    state.transaction_stack.append(
        (MessageCallTransaction(world_state=world_state, gas_limit=8000000), None)
    )
    return state


@patch(
    "mythril.laser.ethereum.instructions.get_call_parameters",
    return_value=(
        "0",
        Account(code=Disassembly(code="0x00"), address="0x19"),
        0,
        0,
        0,
        0,
        0,
    ),
)
def test_staticcall(f1):
    # Arrange
    state = get_global_state()
    state.mstate.stack = [10, 10, 10, 10, 10, 10, 10, 10, 0]
    instruction = Instruction("staticcall", dynamic_loader=None)

    # Act and Assert
    with pytest.raises(TransactionStartSignal) as ts:
        instruction.evaluate(state)
    assert ts.value.transaction.static
    assert ts.value.transaction.initial_global_state().environment.static


test_data = (
    "selfdestruct",
    "create",
    "create2",
    "log0",
    "log1",
    "log2",
    "log3",
    "log4",
    "sstore",
)


@pytest.mark.parametrize("input", test_data)
def test_staticness(input):
    # Arrange
    state = get_global_state()
    state.environment.static = True
    state.mstate.stack = []
    instruction = Instruction(input, dynamic_loader=None)

    # Act and Assert
    with pytest.raises(WriteProtection):
        instruction.evaluate(state)


test_data_call = ((0, True), (100, False))


@pytest.mark.parametrize("input, success", test_data_call)
@patch("mythril.laser.ethereum.instructions.get_call_parameters")
def test_staticness_call_concrete(f1, input, success):
    # Arrange
    state = get_global_state()
    state.environment.static = True
    state.mstate.stack = [10] * 100
    code = Disassembly(code="616263")
    f1.return_value = ("0", Account(code=code, address="0x19"), 0, input, 0, 0, 0)
    instruction = Instruction("call", dynamic_loader=None)

    # Act and Assert
    if success:
        with pytest.raises(TransactionStartSignal) as ts:
            instruction.evaluate(state)
        assert ts.value.transaction.static
    else:
        with pytest.raises(WriteProtection):
            instruction.evaluate(state)


@patch("mythril.laser.ethereum.instructions.get_call_parameters")
def test_staticness_call_symbolic(f1):
    # Arrange
    state = get_global_state()
    state.environment.static = True
    state.mstate.stack = [10] * 100
    call_value = symbol_factory.BitVecSym("x", 256)
    code = Disassembly(code="616263")
    f1.return_value = ("0", Account(code=code, address="0x19"), 0, call_value, 0, 0, 0)
    instruction = Instruction("call", dynamic_loader=None)

    # Act and Assert
    with pytest.raises(TransactionStartSignal) as ts:
        instruction.evaluate(state)

    assert ts.value.transaction.static
    assert ts.value.global_state.world_state.constraints[-1] == (call_value == 0)



from collections import namedtuple
from datetime import datetime
from typing import Dict, List, Tuple

# Type annotations:
#   start_time: datetime
#   end_time: datetime
_InstrExecRecord = namedtuple("_InstrExecRecord", ["start_time", "end_time"])

# Type annotations:
#   total_time: float
#   total_nr: float
#   min_time: float
#   max_time: float
_InstrExecStatistic = namedtuple(
    "_InstrExecStatistic", ["total_time", "total_nr", "min_time", "max_time"]
)

# Map the instruction opcode to its records if all execution times
_InstrExecRecords = Dict[str, List[_InstrExecRecord]]

# Map the instruction opcode to the statistic of its execution times
_InstrExecStatistics = Dict[str, _InstrExecStatistic]


class InstructionProfiler:
    """Performance profile for the execution of each instruction.
    """

    def __init__(self):
        self.records = dict()

    def record(self, op: int, start_time: datetime, end_time: datetime):
        try:
            self.records[op].append(_InstrExecRecord(start_time, end_time))
        except KeyError:
            self.records[op] = [_InstrExecRecord(start_time, end_time)]

    def _make_stats(self) -> Tuple[float, _InstrExecStatistics]:
        periods = {
            op: list(
                map(lambda r: r.end_time.timestamp() - r.start_time.timestamp(), rs)
            )
            for op, rs in self.records.items()
        }

        stats = dict()
        total_time = 0

        for _, (op, times) in enumerate(periods.items()):
            stat = _InstrExecStatistic(
                total_time=sum(times),
                total_nr=len(times),
                min_time=min(times),
                max_time=max(times),
            )
            total_time += stat.total_time
            stats[op] = stat

        return total_time, stats

    def __str__(self):
        total, stats = self._make_stats()

        s = "Total: {} s\n".format(total)

        for op in sorted(stats):
            stat = stats[op]
            s += "[{:12s}] {:>8.4f} %,  nr {:>6},  total {:>8.4f} s,  avg {:>8.4f} s,  min {:>8.4f} s,  max {:>8.4f} s\n".format(
                op,
                stat.total_time * 100 / total,
                stat.total_nr,
                stat.total_time,
                stat.total_time / stat.total_nr,
                stat.min_time,
                stat.max_time,
            )

        return s



from abc import ABC, abstractmethod


class ExecutionInfo(ABC):
    @abstractmethod
    def as_dict(self):
        """Returns a dictionary with the execution info contained in this object

        The returned dictionary only uses primitive types.
        """
        pass



from mythril.laser.ethereum.state.annotation import (
    StateAnnotation,
    MergeableStateAnnotation,
)

from copy import copy
from typing import Dict, List, Set
import logging

log = logging.getLogger(__name__)


class MutationAnnotation(StateAnnotation):
    """Mutation Annotation

    This is the annotation used by the MutationPruner plugin to record mutations
    """

    def __init__(self):
        pass

    @property
    def persist_over_calls(self) -> bool:
        return True


class DependencyAnnotation(MergeableStateAnnotation):
    """Dependency Annotation

    This annotation tracks read and write access to the state during each transaction.
    """

    def __init__(self):
        self.storage_loaded = set()  # type: Set
        self.storage_written = {}  # type: Dict[int, Set]
        self.has_call = False  # type: bool
        self.path = [0]  # type: List
        self.blocks_seen = set()  # type: Set[int]

    def __copy__(self):
        result = DependencyAnnotation()
        result.storage_loaded = copy(self.storage_loaded)
        result.storage_written = copy(self.storage_written)
        result.has_call = self.has_call
        result.path = copy(self.path)
        result.blocks_seen = copy(self.blocks_seen)
        return result

    def get_storage_write_cache(self, iteration: int):
        return self.storage_written.get(iteration, set())

    def extend_storage_write_cache(self, iteration: int, value: object):
        if iteration not in self.storage_written:
            self.storage_written[iteration] = set()
        self.storage_written[iteration].add(value)

    def check_merge_annotation(self, other: "DependencyAnnotation"):
        if not isinstance(other, DependencyAnnotation):
            raise TypeError("Expected an instance of DependencyAnnotation")
        return self.has_call == other.has_call and self.path == other.path

    def merge_annotation(self, other: "DependencyAnnotation"):
        merged_annotation = DependencyAnnotation()
        merged_annotation.blocks_seen = self.blocks_seen.union(other.blocks_seen)
        merged_annotation.has_call = self.has_call
        merged_annotation.path = copy(self.path)
        merged_annotation.storage_loaded = self.storage_loaded.union(
            other.storage_loaded
        )
        keys = set(
            list(other.storage_written.keys()) + list(self.storage_written.keys())
        )
        for key in keys:
            other_set = other.storage_written.get(key, set())
            merged_annotation.storage_written[key] = self.storage_written.get(
                key, set()
            ).union(other_set)
        return merged_annotation


class WSDependencyAnnotation(MergeableStateAnnotation):
    """Dependency Annotation for World state

    This  world state annotation maintains a stack of state annotations.
    It is used to transfer individual state annotations from one transaction to the next.
    """

    def __init__(self):
        self.annotations_stack: List[DependencyAnnotation] = []

    def __copy__(self):
        result = WSDependencyAnnotation()
        result.annotations_stack = copy(self.annotations_stack)
        return result

    def check_merge_annotation(self, annotation: "WSDependencyAnnotation") -> bool:
        if len(self.annotations_stack) != len(annotation.annotations_stack):
            # We can only merge worldstate annotations that have seen an equal amount of transactions
            # since the beginning of symbolic execution
            return False
        for a1, a2 in zip(self.annotations_stack, annotation.annotations_stack):
            if a1 == a2:
                continue
            if (
                isinstance(a1, MergeableStateAnnotation)
                and isinstance(a2, MergeableStateAnnotation)
                and a1.check_merge_annotation(a2) is True
            ):
                continue
            log.debug("Aborting merge between annotations {} and {}".format(a1, a2))
            return False

        return True

    def merge_annotation(
        self, annotation: "WSDependencyAnnotation"
    ) -> "WSDependencyAnnotation":
        merged_annotation = WSDependencyAnnotation()
        for a1, a2 in zip(self.annotations_stack, annotation.annotations_stack):
            if a1 == a2:
                merged_annotation.annotations_stack.append(copy(a1))
            merged_annotation.annotations_stack.append(a1.merge_annotation(a2))
        return merged_annotation



import time
from mythril.support.support_utils import Singleton


class TimeHandler(object, metaclass=Singleton):
    def __init__(self):
        self._start_time = None
        self._execution_time = None

    def start_execution(self, execution_time):
        self._start_time = int(time.time() * 1000)
        self._execution_time = execution_time * 1000

    def time_remaining(self):
        return self._execution_time - (int(time.time() * 1000) - self._start_time)


time_handler = TimeHandler()



from ethereum import opcodes
from ethereum.utils import ceil32
from typing import Callable, Dict, Tuple, Union
from mythril.support.opcodes import OPCODES, STACK, GAS


def calculate_sha3_gas(length: int):
    """

    :param length:
    :return:
    """
    gas_val = 30 + opcodes.GSHA3WORD * (ceil32(length) // 32)
    return gas_val, gas_val


def calculate_native_gas(size: int, contract: str):
    """

    :param size:
    :param contract:
    :return:
    """
    gas_value = 0
    word_num = ceil32(size) // 32
    if contract == "ecrecover":
        gas_value = opcodes.GECRECOVER
    elif contract == "sha256":
        gas_value = opcodes.GSHA256BASE + word_num * opcodes.GSHA256WORD
    elif contract == "ripemd160":
        gas_value = opcodes.GRIPEMD160BASE + word_num * opcodes.GRIPEMD160WORD
    elif contract == "identity":
        gas_value = opcodes.GIDENTITYBASE + word_num * opcodes.GIDENTITYWORD
    else:
        # TODO: Add gas for other precompiles, computation should be shifted to natives.py
        #  as some data isn't available here
        pass
    return gas_value, gas_value


def get_opcode_gas(opcode: str) -> Tuple[int, int]:
    return OPCODES[opcode][GAS]


def get_required_stack_elements(opcode: str) -> int:
    return OPCODES[opcode][STACK][0]



from typing import Dict, Tuple, Union


Z_OPERATOR_TUPLE = (0, 1)
UN_OPERATOR_TUPLE = (1, 1)
BIN_OPERATOR_TUPLE = (2, 1)
T_OPERATOR_TUPLE = (3, 1)
GAS = "gas"
STACK = "stack"
ADDRESS = "address"

# Gas tuple contains (min_gas, max_gas)
# stack tuple contains (no_of_elements_popped, no_of_elements_pushed)

# TODO: Make this more specific when TypedDict supports key re-usage.
OPCODES: Dict = {
    "STOP": {GAS: (0, 0), STACK: (0, 0), ADDRESS: 0x00},
    "ADD": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x01},
    "MUL": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x02},
    "SUB": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x03},
    "DIV": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x04},
    "SDIV": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x05},
    "MOD": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x06},
    "SMOD": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x07},
    "ADDMOD": {GAS: (8, 8), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x08},
    "MULMOD": {GAS: (8, 8), STACK: T_OPERATOR_TUPLE, ADDRESS: 0x09},
    "EXP": {
        GAS: (10, 340),
        STACK: BIN_OPERATOR_TUPLE,
        ADDRESS: 0x0A,
    },  # exponent max 2^32
    "SIGNEXTEND": {GAS: (5, 5), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x0B},
    "LT": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x10},
    "GT": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x11},
    "SLT": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x12},
    "SGT": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x13},
    "EQ": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x14},
    "ISZERO": {GAS: (3, 3), STACK: UN_OPERATOR_TUPLE, ADDRESS: 0x15},
    "AND": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x16},
    "OR": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x17},
    "XOR": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x18},
    "NOT": {GAS: (3, 3), STACK: UN_OPERATOR_TUPLE, ADDRESS: 0x19},
    "BYTE": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x1A},
    "SHL": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x1B},
    "SHR": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x1C},
    "SAR": {GAS: (3, 3), STACK: BIN_OPERATOR_TUPLE, ADDRESS: 0x1D},
    "SHA3": {
        GAS: (
            30,
            30 + 6 * 8,
        ),  # max can be larger, but usually storage location with 8 words
        STACK: BIN_OPERATOR_TUPLE,
        ADDRESS: 0x20,
    },
    "ADDRESS": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x30},
    "BALANCE": {GAS: (700, 700), STACK: UN_OPERATOR_TUPLE, ADDRESS: 0x31},
    "ORIGIN": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x32},
    "CALLER": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x33},
    "CALLVALUE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x34},
    "CALLDATALOAD": {GAS: (3, 3), STACK: UN_OPERATOR_TUPLE, ADDRESS: 0x35},
    "CALLDATASIZE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x36},
    "CALLDATACOPY": {
        GAS: (2, 2 + 3 * 768),  # https://ethereum.stackexchange.com/a/47556
        STACK: (3, 0),
        ADDRESS: 0x37,
    },
    "CODESIZE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x38},
    "CODECOPY": {
        GAS: (2, 2 + 3 * 768),  # https://ethereum.stackexchange.com/a/47556,
        STACK: (3, 0),
        ADDRESS: 0x39,
    },
    "GASPRICE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x3A},
    "EXTCODESIZE": {GAS: (700, 700), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x3B},
    "EXTCODECOPY": {
        GAS: (700, 700 + 3 * 768),  # https://ethereum.stackexchange.com/a/47556
        STACK: (4, 0),
        ADDRESS: 0x3C,
    },
    "EXTCODEHASH": {GAS: (700, 700), STACK: UN_OPERATOR_TUPLE, ADDRESS: 0x3F},
    "RETURNDATASIZE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x3D},
    "RETURNDATACOPY": {GAS: (3, 3), STACK: (3, 0), ADDRESS: 0x3E},
    "BLOCKHASH": {GAS: (20, 20), STACK: UN_OPERATOR_TUPLE, ADDRESS: 0x40},
    "COINBASE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x41},
    "TIMESTAMP": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x42},
    "NUMBER": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x43},
    "DIFFICULTY": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x44},
    "GASLIMIT": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x45},
    "CHAINID": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x46},
    "SELFBALANCE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x47},
    "BASEFEE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x48},
    "POP": {GAS: (2, 2), STACK: (1, 0), ADDRESS: 0x50},
    # assume 1KB memory r/w cost as upper bound
    "MLOAD": {GAS: (3, 96), STACK: UN_OPERATOR_TUPLE, ADDRESS: 0x51},
    "MSTORE": {GAS: (3, 98), STACK: (2, 0), ADDRESS: 0x52},
    "MSTORE8": {GAS: (3, 98), STACK: (2, 0), ADDRESS: 0x53},
    # assume 64 byte r/w cost as upper bound
    "SLOAD": {GAS: (800, 800), STACK: UN_OPERATOR_TUPLE, ADDRESS: 0x54},
    "SSTORE": {GAS: (5000, 25000), STACK: (1, 0), ADDRESS: 0x55},
    "JUMP": {GAS: (8, 8), STACK: (1, 0), ADDRESS: 0x56},
    "JUMPI": {GAS: (10, 10), STACK: (2, 0), ADDRESS: 0x57},
    "PC": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x58},
    "MSIZE": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x59},
    "GAS": {GAS: (2, 2), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x5A},
    "JUMPDEST": {GAS: (1, 1), STACK: (0, 0), ADDRESS: 0x5B},
    "BEGINSUB": {GAS: (2, 2), STACK: (0, 0), ADDRESS: 0x5C},
    "RETURNSUB": {GAS: (5, 5), STACK: (0, 0), ADDRESS: 0x5D},
    "JUMPSUB": {GAS: (10, 10), STACK: (1, 0), ADDRESS: 0x5E},
    # apparently Solidity only allows byte32 as input to the log
    # function. Virtually it could be as large as the block gas limit
    # allows, but let's stick to the reasonable standard here.
    # https://ethereum.stackexchange.com/a/1691
    "LOG0": {GAS: (375, 375 + 8 * 32), STACK: (2, 0), ADDRESS: 0xA0},
    "LOG1": {GAS: (2 * 375, 2 * 375 + 8 * 32), STACK: (3, 0), ADDRESS: 0xA1},
    "LOG2": {GAS: (3 * 375, 3 * 375 + 8 * 32), STACK: (4, 0), ADDRESS: 0xA2},
    "LOG3": {GAS: (4 * 375, 4 * 375 + 8 * 32), STACK: (5, 0), ADDRESS: 0xA3},
    "LOG4": {GAS: (5 * 375, 5 * 375 + 8 * 32), STACK: (6, 0), ADDRESS: 0xA4},
    "CREATE": {GAS: (32000, 32000), STACK: T_OPERATOR_TUPLE, ADDRESS: 0xF0},
    "CREATE2": {
        GAS: (32000, 32000),  # TODO: Make the gas values dynamic
        STACK: (4, 1),
        ADDRESS: 0xF5,
    },
    "CALL": {GAS: (700, 700 + 9000 + 25000), STACK: (7, 1), ADDRESS: 0xF1},
    "CALLCODE": {GAS: (700, 700 + 9000 + 25000), STACK: (7, 1), ADDRESS: 0xF2},
    "RETURN": {GAS: (0, 0), STACK: (2, 0), ADDRESS: 0xF3},
    "DELEGATECALL": {GAS: (700, 700 + 9000 + 25000), STACK: (6, 1), ADDRESS: 0xF4},
    "STATICCALL": {GAS: (700, 700 + 9000 + 25000), STACK: (6, 1), ADDRESS: 0xFA},
    "REVERT": {GAS: (0, 0), STACK: (2, 0), ADDRESS: 0xFD},
    "SELFDESTRUCT": {GAS: (5000, 30000), STACK: (1, 0), ADDRESS: 0xFF},
    "INVALID": {GAS: (0, 0), STACK: (0, 0), ADDRESS: 0xFE},
}

for i in range(1, 33):
    OPCODES[f"PUSH{i}"] = {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x5F + i}

for i in range(1, 17):
    OPCODES[f"DUP{i}"] = {GAS: (3, 3), STACK: (0, 0), ADDRESS: 0x7F + i}
    OPCODES[f"SWAP{i}"] = {GAS: (3, 3), STACK: Z_OPERATOR_TUPLE, ADDRESS: 0x8F + i}

ADDRESS_OPCODE_MAPPING = {}

for opcode, opcode_data in OPCODES.items():
    ADDRESS_OPCODE_MAPPING[opcode_data[ADDRESS]] = opcode
